@@ -35,6 +35,20 @@ extern "C" {
 struct st_mrn_statuses
 {
   long count_skip;
+  long fast_order_limit;
+};
+
+struct st_mrn_ft_info
+{
+  struct _ft_vft *please;
+  grn_ctx *ctx;
+  grn_obj *res;
+  grn_id rid;
+};
+
+struct st_mrn_slot_data
+{
+  grn_id last_insert_rid;
 };
 
 /* functions */
@@ -42,6 +56,11 @@ int mrn_init(void *hton);
 int mrn_deinit(void *hton);
 handler *mrn_handler_create(handlerton *hton, TABLE_SHARE *share, MEM_ROOT *root);
 void mrn_drop_db(handlerton *hton, char *path);
+int mrn_close_connection(handlerton *hton, THD *thd);
+bool mrn_flush_logs(handlerton *hton); 
+float mrn_ft_find_relevance(FT_INFO *handler, uchar *record, uint length);
+float mrn_ft_get_relevance(FT_INFO *handler);
+void mrn_ft_close_search(FT_INFO *handler);
 
 /* handler class */
 class ha_mroonga: public handler
@@ -59,15 +78,26 @@ class ha_mroonga: public handler
   grn_obj **idx_col;
 
   grn_obj *res;
+  grn_obj *res0;
   grn_table_cursor *cur;
+  grn_table_cursor *cur0;
   grn_id row_id;
+  grn_obj *_score;
+
+  st_mrn_ft_info mrn_ft_info;
 
   char **key_min;
   char **key_max;
   int *key_min_len;
   int *key_max_len;
 
+  longlong limit;
+  grn_table_sort_key *sort_keys;
+  int n_sort_keys;
+
   bool count_skip;
+  bool fast_order_limit;
+  bool fast_order_limit_with_index;
 
 public:
   ha_mroonga(handlerton *hton, TABLE_SHARE *share);
@@ -132,9 +162,12 @@ public:
 
   bool get_error_message(int error, String *buf);
 
+  int reset();
+
 private:
   void check_count_skip(key_part_map start_key_part_map,
                         key_part_map end_key_part_map, bool fulltext);
+  void check_fast_order_limit();
   void store_fields_from_primary_table(uchar *buf, grn_id rid);
 };
 
